/*
 * Copyright 2010-2012 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.junit3;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;
import org.callbackparams.combine.reflect.CallbackRecordsFactory;
import org.callbackparams.internal.template.TestrunCallbacks;

/**
 * Parameterized tests for JUnit-3.8.x can be created by subclassing this class
 * instead of {@link junit.framework.TestCase}.
 * <br/>Upgrade to JUnit-4.x and usage of
 * {@link org.callbackparams.junit4.CallbackParamsRunner CallbackParamsRunner}
 * is strongly recommended! However, that does not necessarily mean quit
 * using the JUnit-3.x API since JUnit-4.x is backward compatible and
 * {@link org.callbackparams.junit4.CallbackParamsRunner CallbackParamsRunner}
 * also can run JUnit-3.x tests.
 * <br/><br/>
 * Thou somewhat ugly, having the test-class subclass this class does not
 * conflict with using
 * {@link org.callbackparams.junit4.CallbackParamsRunner CallbackParamsRunner}
 * to run the test. The only differences would be occasionally better
 * performance and that only
 * {@link org.callbackparams.ParameterizedCallback @ParameterizedCallback}-annotated
 * fields are callback-injected while the alternative combine API that is
 * unlocked by overriding
 * {@link #getCallbackRecords()} or {@link #getCallbackRecordsFactory()}
 * would still be available.
 *
 * @author Henrik Kaipe
 */
public abstract class CallbackTestCase extends TestCase {

    private static final Collection dummyRecord = Arrays.asList(new Object[0]);

    /**
     * Instantiation of rebyted test-class will fail if no callback-record is
     * set right away.
     */
    {
        if (isRebytedTestClass()
                && null == TestrunCallbacks.getCurrentCallbackRecord()) {
            TestrunCallbacks.setCallbackRecord(dummyRecord);
        }
    }

    protected CallbackTestCase(String name) {super(name);}
    protected CallbackTestCase() {}

    private boolean isRebytedTestClass() {
        return TestrunCallbacks.class.isInstance(this);
    }

    private boolean isTestrun() {
        return isRebytedTestClass()
                && dummyRecord != TestrunCallbacks.getCurrentCallbackRecord();
    }

    /**
     * Dummy test-method that does nothing. When the test-runner tries to
     * run this test it is intercepted by the callback framework and
     * test-methods with callback parameters are run instead.
     */
    final public void testSpecialDummyTestMethodThatIsUsedToTrickJunitOnly() {
        throw new AssertionError("This method is not supposed to be called!");
    }
    
    private boolean isInterceptionOfCallbackTestRun() {
        return "testSpecialDummyTestMethodThatIsUsedToTrickJunitOnly"
                .equals(getName());
    }

    private Class rebyteTestClass(
            final Map/*String,String*/ newMethodNameMappings) {
        return CallbackTestSuite.rebyteTestClass(
                getClass(), newMethodNameMappings);
    }

    public void run(final TestResult result) {

        if (isTestrun()) {
            super.run(result);

        } else if (isRebytedTestClass()) {
            if (isInterceptionOfCallbackTestRun()) {
                /* This can occur if a CallbackTestCase-test has been annotated
                 * with @RunWith(CallbackParamsRunner.class), which happens to
                 * be the recommended way to run tests.
                 * It happens when the test-runner tries to run the test-method
                 * testSpecialDummyTestMethodThatIsUsedToTrickJunitOnly(), which
                 * for known reasons is not a test-method to be run ...
                 */
                return;
            }
            TestInstanceFactory factory = new TestInstanceFactory(getClass());
            for (Iterator i = getCombinedRecords().iterator(); i.hasNext() ;) {
                final Object testRunCallbackRecord = i.next();

                try {
                    factory.newTestrun(getName(), testRunCallbackRecord)
                            .run(result);
                } finally {
                    TestrunCallbacks.setCallbackRecord(dummyRecord);
                }
            }

        } else if (isInterceptionOfCallbackTestRun()) {
            Map/*String,String*/ newTestNamesMap = new HashMap();
            final Class rebytedTestClass = rebyteTestClass(newTestNamesMap);

            for (Iterator i = newTestNamesMap.entrySet().iterator()
                    ; i.hasNext() ;) {
                final Map.Entry/*String,String*/ proxyMethodEntr =
                        (Map.Entry) i.next();

                String proxyMethodTestName =
                        TestInstanceFactory.asTestName(proxyMethodEntr);
                TestSuite.createTest(rebytedTestClass, proxyMethodTestName)
                        .run(result);
            }

        } else {
            TestSuite.createTest(rebyteTestClass(null), getName()).run(result);
        }
    }

    public int countTestCases() {
        final int countBase = super.countTestCases();

        if (isTestrun()) {
            return countBase;

        } else if (isRebytedTestClass()) {
            return countBase * getCallbackRecords().size();

        } else if (isInterceptionOfCallbackTestRun()) {
            Map tmp = new HashMap();
            rebyteTestClass(tmp);
            return countBase * tmp.size() * getCallbackRecords().size();

        } else {
            return countBase * getCallbackRecords().size();
        }
    }

    /**
     * Returns a collection of callback-records, i.e. Object[] or Collection
     * instances that contain all callback-values for a particular test-run.
     * This method can be overridden to test specific callback-records.
     * <br/>
     * The default implementation uses the
     * {@link CallbackRecordsFactory} supplied by the overridable method
     * {@link #getCallbackRecordsFactory()} to create the callback-records.
     * <br/><br/>
     * Please note that overriding this method is a non-static and JDK-1.4
     * compatible alternative to the using the annotation
     * {@link org.callbackparams.combine.annotation.CallbackRecords @CallbackRecords}.
     *
     * @see org.callbackparams.combine.annotation.CallbackRecords @CallbackRecords
     */
    protected Collection getCallbackRecords() {
        return getCallbackRecordsFactory()
                .collectCallbackRecordsReflectively(getClass());
    }

    /**
     * Acts as a private wrapper around {@link #getCallbackRecords()} to wrap
     * record elements that have been manually combined by wrapping them with
     * the wrapper class {@link org.callbackparams.combine.reflect.Combined}.
     */
    private Collection getCombinedRecords() {
        Collection callbackRecords = getCallbackRecords();
        Collection combinedRecords = new ArrayList(callbackRecords.size());
        for (Iterator iRecord = callbackRecords.iterator(); iRecord.hasNext();){
            final Object record = iRecord.next();
            combinedRecords.add(CallbackRecordsFactory.asCombinedArray(record));
        }
        return combinedRecords;
    }

    /**
     * Returns the {@link CallbackRecordsFactory} instance that will be used
     * by the default implementation of
     * {@link #getCallbackRecords()} to create the callback-records.
     * The default CallbackRecordsFactory instance returned is the one supplied
     * by the factory method {@link CallbackRecordsFactory#getInstance()}. If
     * this default factory won't fit your purposes you can
     * override this method.
     * <br/>
     * Overriding this method is a particularly interesting alternative if
     * you are stuck with JDK-1.4 and therefore cannot use annotations to
     * instruct the default factory. E.g. you could then have your factory
     * implementation override the method
     * {@link CallbackRecordsFactory#retrieveCombineStrategy(Class)} and let
     * it return a {@link org.callbackparams.combine.CombineCompletely}
     * instance instead of the default
     * {@link org.callbackparams.combine.CombineAllPossible2Combinations}
     * instance.
     * <br/><br/>
     * Overriding this method is a non-static and JDK-1.4 compatible alternative
     * to configuring a combine-strategy by using the annotation
     * {@link org.callbackparams.combine.annotation.CombineConfig @CombineConfig}.
     * An interesting JDK-1.4 oppertunity is to override
     * {@link CallbackRecordsFactory#isCallbackValuesFactoryClass(java.lang.Class)}
     * and
     * {@link CallbackRecordsFactory#retrieveCombinedArray(java.lang.Class)},
     * e.g. to make it recognize Enum-classes a'la commons-lang ...
     *
     * @see CallbackRecordsFactory#retrieveCombineStrategy(java.lang.Class)
     * @see CallbackRecordsFactory#retrieveCombinedArray(java.lang.Class)
     * @see CallbackRecordsFactory#isCallbackValuesFactoryClass(java.lang.Class)
     * @see org.callbackparams.combine.CombineStrategy CombineStrategy
     * @see org.callbackparams.combine.annotation.CombineConfig @CombineConfig
     */
    protected CallbackRecordsFactory getCallbackRecordsFactory() {
        return CallbackRecordsFactory.getInstance();
    }
}
