/*
 * Copyright 2012 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.support.annotation;

import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import org.callbackparams.support.ExceptionUtil;

/**
 * @author Henrik Kaipe
 */
public class AnnotationInjector
extends org.callbackparams.support.AnnotationInjector {

    @Override
    protected void injectAnnotationsInternal(Object target, Object enumConstant)
    {
        Field enumField = findEnumField(enumConstant);
        if (null == enumField) {
            return;
        }
        Class<?> enumClass = enumField.getDeclaringClass();
        Package enumPackage = enumClass.getPackage();
        AnnotatedElement[] annotatedElements = null == enumPackage
                ? new AnnotatedElement[]{enumField, enumClass}
                : new AnnotatedElement[]{enumField, enumClass, enumPackage};

        Collection<Field> injectableFields =
                findInjectableField(target.getClass());
        checkForClassRetention(injectableFields);
        for (Field f : injectableFields) {
            Annotation a = findAnnotation(f.getType(), annotatedElements);

            if (null != a) {
                try {
                    f.setAccessible(true);
                } catch (SecurityException ignoreAndContinueAnyway___) {}
                try {
                    f.set(target, a);
                } catch (Exception ex) {
                    throw ExceptionUtil.unchecked(ex);
                }
            }
        }
    }

    private static Annotation findAnnotation(
            Class<?> annotationClass, AnnotatedElement[] elements) {
        Class<? extends Annotation> c =
                annotationClass.asSubclass(Annotation.class);
        for (AnnotatedElement ae : elements) {
            Annotation a = ae.getAnnotation(c);
            if (null != a) {
                return a;
            }
        }
        return null;
    }

    private static Field findEnumField(Object enumConstant) {
        Class<? extends Enum> c = findEnumSuperClass(enumConstant.getClass());
        if (null == c) {
            return null;
        } else {
            try {
                return c.getField(((Enum) enumConstant).name());
            } catch (Exception ex) {
                throw ExceptionUtil.unchecked(ex);
            }
        }
    }

    private static Class<? extends Enum> findEnumSuperClass(Class<?> c) {
        if (c.isEnum()) {
            return c.asSubclass(Enum.class);
        } else if (Object.class == c) {
            return null;
        } else {
            return findEnumSuperClass(c.getSuperclass());
        }
    }

    private static Collection<Field> findInjectableField(Class<?> targetClass) {
        if (Object.class == targetClass) {
            return new ArrayList<Field>();
        }

        Collection<Field> injectableFields =
                findInjectableField(targetClass.getSuperclass());
        for (Field f : targetClass.getDeclaredFields()) {
            if (isNonStaticAndNonFinal(f) && isAnnotationClass(f.getType())) {
                injectableFields.add(f);
            }
        }
        return injectableFields;
    }

    private static boolean isNonStaticAndNonFinal(Field f) {
        int modifiers = f.getModifiers();
        return false == Modifier.isStatic(modifiers)
                && false == Modifier.isFinal(modifiers);
    }

    private static boolean isAnnotationClass(Class<?> c) {
        Retention retention = c.getAnnotation(Retention.class);
        return null == retention
                ? c.isAnnotation()
                : RetentionPolicy.RUNTIME == retention.value();
    }

    private static void checkForClassRetention(Collection<Field> injectableFields) {
        List<Field> fieldsWithClassRetention = new ArrayList<Field>();
        for (Field f : injectableFields) {
            if (false == f.getType().isAnnotationPresent(Retention.class)) {
                fieldsWithClassRetention.add(f);
            }
        }
        if (false == fieldsWithClassRetention.isEmpty()) {
            throw new Error("Annotation classes that are not annotated with "
                    + "@Retention(RetentionPolicy.RUNTIME) cannot injected: "
                    + fieldsWithClassRetention);
        }
    }
}
