/*
 * Copyright 2011-2012 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.wrap.legacy;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import org.apache.bcel.classfile.EmptyVisitor;
import org.apache.bcel.classfile.Visitor;
import org.callbackparams.support.ClassConstantModifyingVisitor;
import org.callbackparams.support.ClassWrapper;
import org.callbackparams.support.ExceptionUtil;
import org.callbackparams.support.TemplateBasedClassBuilder;

/**
 * @author Henrik Kaipe
 */
public class WrappingLegacySupport extends WrappingToStringTemplate {

    /**
     * Is used to make the static template field
     * {@link WrappingTemplate#mostRecentlyInjectedWrappedValue}
     * easily accessable when the template is used.
     */
    private static Visitor makeStaticFieldPublic = new EmptyVisitor() {
        public void visitField(org.apache.bcel.classfile.Field field) {
            if (field.isStatic()) {
                field.isPublic(true);
            }
        }
    };

    public static Wrapping[] wrapArray(Class targetClass, Object[] wrappedValues) {
        final Wrapping[] result = (Wrapping[])
                Array.newInstance(targetClass, wrappedValues.length);
        final ClassWrapper wrappedTargetClass = wrapClass(targetClass);
        for (int i = 0; i < result.length; ++i) {
            result[i] = wrap(wrappedTargetClass, wrappedValues[i]);
        }
        return result;
    }

    public static Wrapping wrapSingle(Class targetClass, Object wrappedValue) {
        return wrap(wrapClass(targetClass), wrappedValue);
    }

    private static Wrapping wrap(ClassWrapper wrappedClass, Object wrappedValue){
        ThreadLocal injectionField =
                findInjectionField(wrappedClass.getWrappedClass());
        Object backupOfPossiblyPendingWrappedValueThatWasInjectedEarlierInStack=
                injectionField.get();
        injectionField.set(wrappedValue);

        Object actualValue = null;
        try {
            Wrapping w = (Wrapping) wrappedClass.newInstance();
            actualValue = w.wrappedValue();
            if (wrappedValue != actualValue) {
                throw new Error("Tried to wrap " + wrappedValue
                        + "; but actually wrapped value was: " + actualValue);
            }
            return w;
        } finally {
            actualValue = injectionField.get();
            injectionField.set(
                    backupOfPossiblyPendingWrappedValueThatWasInjectedEarlierInStack);
            if (wrappedValue != actualValue) {
                System.err.println(/* Should help debuging when needed: */
                        "Value injected for wrapping was " + wrappedValue
                        + "; but actually injected value was:" + actualValue);
            }
        }
    }

    private static ThreadLocal findInjectionField(Class wrappingClass) {
        if (WrappingTemplate.class.isAssignableFrom(wrappingClass)) {
            return WrappingTemplate.mostRecentlyInjectedWrappedValue;
        } else {
            String wrappingTemplateClassName = WrappingTemplate.class.getName();
            while (false ==
                    wrappingClass.getName().equals(wrappingTemplateClassName)) {
                wrappingClass = wrappingClass.getSuperclass();
            }
            final Field[] fields = wrappingClass.getDeclaredFields();
            for (int i = 0; i < fields.length; ++i) {
                Field f = fields[i];
                if (Modifier.isStatic(f.getModifiers())
                        && ThreadLocal.class == f.getType()) {
                    f.setAccessible(true);
                    try {
                        return (ThreadLocal) f.get(null);
                    } catch (Exception ex) {
                        throw ExceptionUtil.unchecked(ex);
                    }
                }
            }
        }
        throw new Error("Unable to find the ThreadLocal injection field on "
                + wrappingClass);
    }

    private static ClassWrapper wrapClass(Class targetClass) {
        if (targetClass.isInterface()
                || false == Wrapping.class.isAssignableFrom(targetClass)) {
            throw new NotAWrappingClassException(targetClass);
        }
        if (null != targetClass.getDeclaringClass()
                && false == Modifier.isStatic(targetClass.getModifiers())) {
            throw new NonStaticInnerWrappingClassException(targetClass);
        }

        if (false == Modifier.isAbstract(targetClass.getModifiers())) {
            return new ClassWrapper(targetClass);
        }

        boolean extendWithToStringTemplate =
                lacksToStringImplementation(targetClass);
        ClassConstantModifyingVisitor newWrappingTemplateSuperClass =
                new ClassConstantModifyingVisitor(
                DummySuperClass.class, targetClass);
        ClassWrapper targetClassSubClass = TemplateBasedClassBuilder
                .newBuilder(targetClass.getClassLoader(), WrappingTemplate.class)
                .applyVisitor(newWrappingTemplateSuperClass)
                .applyFieldVisitor(makeStaticFieldPublic)
                .wrapClass(extendWithToStringTemplate);
        if (false == extendWithToStringTemplate) {
            return targetClassSubClass;
        } else {
            return TemplateBasedClassBuilder.newBuilder(
                    targetClassSubClass.getWrappedClass().getClassLoader(),
                    WrappingToStringTemplate.class)
                    .wrapClass();
        }
    }

    private static boolean lacksToStringImplementation(Class targetClass) {
        try {
            Method toStringMethod = targetClass.getMethod("toString", null);
            return Modifier.isAbstract(toStringMethod.getModifiers())
                    || false == Wrapping.class
                    .isAssignableFrom(toStringMethod.getDeclaringClass());
        } catch (Exception shouldNeverHappen) {
            throw ExceptionUtil.unchecked(shouldNeverHappen);
        }
    }
}
